package droidface

import soot.{Unit => SootUnit, Value => SootValue, util => _, _}
// import scala.collection.mutable.{Buffer, Queue, Map}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._
import Lattice.Level._
import soot.util.{JasminOutputStream}
import soot.options.Options

import java.io.{PrintWriter, OutputStreamWriter, FileOutputStream}

import scala.collection.JavaConverters._

// TODO: check if inheriting from SootClass causing any Scala-Java interoperability
// problems and do that if not.
class FacetedClass(t: Type, sootClass: SootClass) {
  val objClass = Scene.v.loadClassAndSupport("java.lang.Object")
  val sootClasss = sootClass
  sootClass.setSuperclass(objClass)
  sootClass.setModifiers(Modifier.PUBLIC)
  var levelFields: Map[Level, SootFieldRef] = Map()
  Lattice.levels.toSeq map { l =>
    val field = new SootField(l.toString, t, Modifier.PUBLIC)
    sootClass.addField(field)
    levelFields += ((l, field.makeRef))
  }
  // Add a default constructor:
  val constr: SootMethod = new SootMethod("<init>", Seq().asJava, VoidType.v,
    Modifier.PUBLIC)
  val constrBody = Jimple.v.newBody(constr)
  constr.setActiveBody(constrBody)
  val lg = new soot.javaToJimple.LocalGenerator(constrBody)
  val thisLoc = lg.generateLocal(sootClass.getType)
  constrBody.getUnits.addFirst(
    Jimple.v.newIdentityStmt(thisLoc,
      Jimple.v.newThisRef(sootClass.getType)))
  sootClass.addMethod(constr)
  val ref = Jimple.v.newParameterRef(sootClass.getType, 0)
  val objConstr = objClass.getMethods.asScala.find(_.isConstructor()).get
  constrBody.getUnits.addLast(Jimple.v.newInvokeStmt(
    Jimple.v.newSpecialInvokeExpr(thisLoc, objConstr.makeRef)))
  constrBody.getUnits.addLast(Jimple.v.newReturnVoidStmt)
  Scene.v.addClass(sootClass)
  // Util.debug(s"Added $sootClass with fields: ${newClass.getFields}")
  val fn = SourceLocator.v.getFileNameFor(sootClass, Options.output_format_class)
  val streamOut = new JasminOutputStream(new FileOutputStream(fn))
  val writerOut = new PrintWriter(new OutputStreamWriter(streamOut))
  new JasminClass(sootClass).print(writerOut)
  writerOut.flush()
  streamOut.close()

  def getLevelField(l: Level): SootFieldRef =
    levelFields.getOrElse(l, Util.die(s"No field for level $l"))

  def getSootClass: SootClass = sootClasss

  def getConstructor: SootMethod = constr
}
