package droidface

import soot.{Unit => SootUnit, Value => SootValue, util => _, _}
// import scala.collection.mutable.{Buffer, Queue, Map}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._
import Lattice.Level._

import scala.collection.JavaConverters._

trait FacetedVar {
  def levelVal(l: Level): SootValue
  def getType: Type
}

class FacetedLocal(local: Local, body: Body) extends FacetedVar {
  val name = TaintTransformer.freshVarCopy(local.getName(),
    (body.getLocals.asScala map (_.toString())) toSet)
  val valueMap: Map[Level, Local] =
    Lattice.levels.map (l => (l, Jimple.v.newLocal(name + l.toString,
      local.getType))) toMap
  val typ: Type = local.getType
  valueMap.values foreach (body.getLocals.add(_))

  override def levelVal(l: Level) =
    valueMap.getOrElse(l, sys.error(s"Bug: no copy of $name for level $l"))
  override def getType: Type = typ
}

class FacetedConst(v: SootValue, t: Type) extends FacetedVar {
  val const = v
  val typ = t

  def this(v: SootValue) = this(v, v.getType)

  override def levelVal(l: Level) = const
  override def getType: Type = typ
}

class FacetedField(fld: SootField, cls: SootClass) {
  val name = TaintTransformer.freshVarCopy(fld.getName,
    cls.getFields.asScala map (_.getName) toSet)
  val typ: Type = fld.getType
  val valueMap: Map[Level, SootFieldRef] =
    Lattice.levels map { l =>
      
      val lfld = new SootField(name + l.toString, typ)
      cls.addField(lfld)
      lfld.setModifiers(fld.getModifiers)
      (l, lfld.makeRef)
    } toMap

  def levelRef(l: Level): SootFieldRef =
    valueMap.getOrElse(l, sys.error(s"Bug: no copy of $name for level $l"))
  def getType: Type = typ
}
