package droidface

import soot.toolkits.graph._
import soot.Body
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._
import java.io.{File, FileOutputStream, PrintWriter}
import sys.process._
import scala.util.matching._
import scalatags.Text.all._
import soot.{Unit => SootUnit, util => _, _}

//import scala.collection._
//import scala.collection.immutable._

object Util {
  var sound: Boolean = true
  var overview: Overview = new Overview()
  var indent: Integer = 0
  var nsu: Boolean = true

  def debug(msg: Object): Unit = println(s"${"  " * Util.indent}$msg")
  
  def warn(msg: Object): Unit = debug(s"WARN: $msg")

  def error(msg: Object): Unit = debug(s"ERR: $msg")

  def makeDotGraph[N](cfg: UnitGraph): DotGraph = {
    val conv = new CFGToDotGraph()
    conv.drawCFG(cfg, cfg.getBody)
  }
  
  def renderDotGraph(label: String = "", dg: DotGraph): String = {
    var tempFile = File.createTempFile("_taintifc", ".dot")
    var outStream = new FileOutputStream(tempFile)
    // dg.setGraphLabel(label)
    dg.setGraphLabel(label)
    dg.render(outStream, 0)
    outStream.close()
    var retCode = Seq("dot", "-Tpng", "-O", tempFile.getAbsolutePath).!
    if (retCode != 0) {
      Util.die("Running dot failed!")
    } else {
      tempFile.getAbsolutePath ++ ".png"
    }
  }

  def renderCFG(cfg: UnitGraph, label:String = ""): String = {
    renderDotGraph(label, makeDotGraph(cfg))
  }

  def die(o: Object, exitCode: Integer = 2): Nothing = {
    error(o)
    // Throw something so we get a stack trace
    // TODO: use exceptions properly
    throw new IllegalAccessError("")
    System.exit(exitCode)
    ???
  }

  def setIndent(i: Integer): Unit = { Util.indent = i }

  def incIndent(): Unit = setIndent(indent + 1)

  def decIndent(): Unit = setIndent(scala.math.max(indent - 1, 0))

  def indented(act: =>Unit): Unit = {
    incIndent()
    act
    decIndent()
  }
  
  def unsound(): Unit = { sound = false }
}
